@@ -362,16 +362,21 @@
                    '((0 . "Escaped non-special character `}'")
                      (4 . "Escaped non-special character `a'")
                      (8 . "Escaped non-special character `%'"))))
-    (should (equal (xr-lint "a?+b+?\\(?:c?\\)*d\\{3\\}+e*?\\{2,5\\}")
-                   '((2  . "Repetition of repetition")
+    (should (equal (xr-lint "a?+b+?\\(?:c*\\)*d\\{3\\}+e*?\\{2,5\\}")
+                   '((2  . "Repetition of option")
                      (14 . "Repetition of repetition")
                      (25 . "Repetition of repetition"))))
     (should (equal (xr-lint "\\(a*\\)*\\(b+\\)*\\(c*\\)?\\(d+\\)?")
                    '((6 .  "Repetition of repetition")
                      (13 . "Repetition of repetition")
-                     (20 . "Repetition of repetition"))))
+                     (20 . "Optional repetition"))))
+    (should (equal (xr-lint "\\(a?\\)+\\(b?\\)?")
+                   '((6 . "Repetition of option")
+                     (13 . "Optional option"))))
     (should (equal (xr-lint "\\(e*\\)\\{3\\}")
                    '((6 . "Repetition of repetition"))))
+    (should (equal (xr-lint "\\(a?\\)\\{4,7\\}")
+                   '((6 . "Repetition of option"))))
     (should (equal (xr-lint "[]-Qa-fz-t]")
                    '((1 . "Reversed range `]-Q' matches nothing")
                      (7 . "Reversed range `z-t' matches nothing"))))
@@ -404,15 +409,23 @@
              (xr-lint "[-A-Z][A-Z-][A-Z-a][^-A-Z][]-a][A-Z---.]")
              '((16 .
                 "Literal `-' not first or last in character alternative"))))
+    (should (equal (xr-lint "\\'*\\<?\\(?:$\\)+")
+                   '((2 . "Repetition of zero-width assertion")
+                     (5 . "Optional zero-width assertion")
+                     (13 . "Repetition of zero-width assertion"))))
+    ))
+
+(ert-deftest xr-lint-repetition-of-empty ()
+  (let ((text-quoting-style 'grave))
     (should (equal
              (xr-lint "\\(?:a*b?\\)*\\(c\\|d\\|\\)+\\(^\\|e\\)*\\(?:\\)*")
              '((10 . "Repetition of expression matching an empty string")
                (21 . "Repetition of expression matching an empty string"))))
-    (should (equal (xr-lint "\\'*\\<?\\(?:$\\)+")
-                   '((2 . "Repetition of zero-width assertion")
-                     (5 . "Repetition of zero-width assertion")
-                     (13 . "Repetition of zero-width assertion"))))
-    ))
+    (should (equal
+             (xr-lint "\\(?:a*?b??\\)+?")
+             '((12 . "Repetition of expression matching an empty string"))))
+    (should (equal (xr-lint "\\(?:a*b?\\)?")
+                   '((10 . "Optional expression matching an empty string"))))))
 
 (ert-deftest xr-lint-branch-subsumption ()
   (let ((text-quoting-style 'grave))
@@ -509,6 +522,14 @@
     (should (equal (xr-lint "\\Ca\\|ü")
                    '((5 . "Branch matches subset of a previous branch"))))
     
+    (should (equal (xr-lint "\\w\\|[^z-a]")
+                   '((4 . "Branch matches superset of a previous branch"))))
+    (should (equal (xr-lint "\\W\\|[^z-a]")
+                   '((4 . "Branch matches superset of a previous branch"))))
+    (should (equal (xr-lint "\\w\\|a")
+                   '((4 . "Branch matches subset of a previous branch"))))
+    (should (equal (xr-lint "\\W\\|\f")
+                   '((4 . "Branch matches subset of a previous branch"))))
   ))
 
 (ert-deftest xr-lint-subsumed-repetition ()
@@ -592,6 +613,60 @@
                      (46 . "Repetition subsumed by preceding repetition"))))
     ))
 
+(ert-deftest xr-lint-wrapped-subsumption ()
+  (let ((text-quoting-style 'grave))
+    (should (equal
+             (xr-lint "\\(?:a*x[ab]+\\)*")
+             '((14 . "Last item in repetition subsumes first item (wrapped)"))))
+    (should (equal
+             (xr-lint "\\([ab]*xya?\\)+")
+             '((13 . "First item in repetition subsumes last item (wrapped)"))))
+    (should (equal
+             (xr-lint "\\(?3:a*xa*\\)\\{7\\}")
+             '((14 . "Last item in repetition subsumes first item (wrapped)"))))
+    ))
+
+(ert-deftest xr-lint-bad-anchor ()
+  (let ((text-quoting-style 'grave))
+    (should (equal (xr-lint "a\\(?:^\\)")
+                   '((1 . "Non-newline followed by line-start anchor"))))
+    (should (equal (xr-lint "a?\\(?:^\\)")
+                   '((2 . "Non-newline followed by line-start anchor"))))
+    (should (equal (xr-lint "a\\(?:^\\|b\\)")
+                   '((1 . "Non-newline followed by line-start anchor"))))
+    (should (equal (xr-lint "a?\\(?:^\\|b\\)")
+                   nil))
+    (should (equal (xr-lint "\\(?:$\\)a")
+                   '((7 . "End-of-line anchor followed by non-newline"))))
+    (should (equal (xr-lint "\\(?:$\\)\\(\n\\|a\\)")
+                   '((7 . "End-of-line anchor followed by non-newline"))))
+    (should (equal (xr-lint "\\(?:$\\|b\\)a")
+                   '((10 . "End-of-line anchor followed by non-newline"))))
+    (should (equal (xr-lint "\\(?:$\\|b\\)\\(\n\\|a\\)")
+                   nil))
+    (should (equal (xr-lint ".\\(?:^$\\).")
+                   '((1 . "Non-newline followed by line-start anchor")
+                     (9 . "End-of-line anchor followed by non-newline"))))
+    (should (equal (xr-lint "\\'b")
+                   '((2 . "End-of-text anchor followed by non-empty pattern"))))
+    (should (equal (xr-lint "\\'b?")
+                   '((3 . "End-of-text anchor followed by non-empty pattern"))))
+    (should (equal (xr-lint "\\(?:a\\|\\'\\)b")
+                   '((11 .
+                      "End-of-text anchor followed by non-empty pattern"))))
+    (should (equal (xr-lint "\\(?:a\\|\\'\\)b?")
+                   nil))
+    ))
+
+(ert-deftest xr-lint-file ()
+  (let ((text-quoting-style 'grave))
+    (should (equal (xr-lint "a.b\\.c.*d.?e.+f." 'file)
+                   '((1 . "Possibly unescaped `.' in file-matching regexp")
+                     (15 . "Possibly unescaped `.' in file-matching regexp"))))
+    (should (equal (xr-lint "^abc$" 'file)
+                   '((0 . "Use \\` instead of ^ in file-matching regexp")
+                     (4 . "Use \\' instead of $ in file-matching regexp"))))))
+
 (ert-deftest xr-skip-set ()
   (should (equal (xr-skip-set "0-9a-fA-F+*")
                  '(any "0-9a-fA-F" "+*")))
